import numpy as np


STANDARD_SIZE = np.array([4096 * 2, 5120 * 2, 2044])
STANDARD_MIN_VALUES = np.array([-4096, -5120, 0])
STANDARD_MAX_VALUES = np.array([4096, 5120, 2000])


class LocationSplitManager:
    def __init__(self, num_layers, layer_grid_sizes, min_values, max_values):
        """
        Creates a location split manager
        :param num_layers: Number of layers for this grid.
        :param layer_grid_sizes:
        :param total_size:
        """
        self.min_values = min_values
        self.max_values = max_values
        self.layer_grid_sizes = layer_grid_sizes
        self.num_layers = num_layers
        self.total_size = max_values - min_values

    def create_boxes(self, position):
        output_tiles = []
        for i in range(self.num_layers):
            layer_mult = (1 + i)
            max_layer_value = (self.layer_grid_sizes ** layer_mult)
            raw_scaled_position = ((position - self.min_values) /
                                   self.total_size) * max_layer_value
            print(raw_scaled_position)
            full_position = np.floor(np.maximum(0, np.minimum(raw_scaled_position, max_layer_value - 1)))

            output_tiles.append(full_position % self.layer_grid_sizes)
        return output_tiles



import inspect
import logging
import os
import subprocess
import json
import platform as pt
from typing import List

from carball.rattletrap.rattletrap_utils import get_rattletrap_binaries, download_rattletrap, get_rattletrap_path, \
    get_binary_for_platform

logger = logging.getLogger(__name__)


def create_rattletrap_command(replay_path: str, output_path: str, overwrite: bool=True) -> List[str]:
    """
    Takes a path to the replay and outputs the json of that replay.

    :param replay_path: Path to a specific replay.
    :param output_path: The output path of rattletrap.
    :param overwrite: True if we should recreate the json even if it already exists.
    :return: The json created from rattle trap.
    """
    rattletrap_path = get_rattletrap_path()
    binaries = get_rattletrap_binaries(rattletrap_path)
    if len(binaries) == 0:
        logger.warning("Need to redownload rattletrap")
        download_rattletrap()
        binaries = get_rattletrap_binaries(rattletrap_path)
    binary = get_binary_for_platform(pt.system(), binaries)
    if binary is None:
        print('no binary!')
    cmd = [os.path.join(rattletrap_path, '{}'.format(binary)), '--compact', '-i',
           replay_path]
    if output_path:
        output_dirs = os.path.dirname(output_path)
        if not os.path.isdir(output_dirs) and output_dirs != '':
            os.makedirs(output_dirs)
        if overwrite or not os.path.isfile(output_path):
            cmd += ['--output', output_path]

    return cmd


def run_rattletrap_command(command: List[str], output_path: str):
    proc = subprocess.Popen(command, stdout=subprocess.PIPE)
    output = proc.stdout.read()
    if output:
        output = output.decode('utf8')
    if output_path:
        with open(output_path, encoding="utf8") as f:
            _json = json.load(f)
    else:
        _json = json.loads(output)
    return _json


def decompile_replay(replay_path: str, output_path: str, overwrite: bool=True):
    """
    Takes a path to the replay and outputs the json of that replay.

    :param replay_path: Path to a specific replay.
    :param output_path: The output path of rattletrap.
    :param overwrite: True if we should recreate the json even if it already exists.
    :return: The json created from rattle trap.
    """
    command = create_rattletrap_command(replay_path, output_path, overwrite)
    logger.debug(" ".join(command))
    return run_rattletrap_command(command, output_path)



from typing import ClassVar, List, Optional

from ..base_checks.player_check import PlayerCheck


class PlayerDataFrameCheck(PlayerCheck):
    class_message: ClassVar[str] = "All columns in player data frame must have at least 1 truthy value."

    included_columns = ['pos_x', 'pos_y', 'pos_z']

    def critical_checks(self) -> List[Optional[AssertionError]]:
        player = self.obj
        if player.data is not None:
            return [
                self.check(condition, f'player.data column: {column_name} has truthy values')
                for column_name, condition in player.data.any(axis=0).iteritems()
                if column_name in self.included_columns
            ]
        else:
            return [
                self.check(False, 'player has non-None .data attribute')
            ]



import logging

import pandas as pd

from ...analysis.cleaner.frame_cleaner import remove_invalid_players
from ...generated.api import game_pb2
from ...json_parser.game import Game


logger = logging.getLogger(__name__)


def clean_replay(game: Game, data_frame: pd.DataFrame, proto_game: game_pb2.Game, player_map):
    drop_nans(data_frame)

    remove_invalid_players(game, data_frame, proto_game, player_map)

    logger.info('cleaned up the replay')


def drop_nans(data_frame: pd.DataFrame):
    data_frame.dropna(axis=1, how='all')
    post_hit = data_frame[(data_frame.game.time > 5.0)]
    columns_to_remove = post_hit.columns[post_hit.isna().all()].tolist()
    if len(columns_to_remove) > 0:
        data_frame.drop(columns_to_remove, axis=1, inplace=True)
        logger.warning('Dropping these columns' + str(columns_to_remove))



from typing import List, Optional

from ..base_checks.game_check import GameCheck


class BallDataFrameCheck(GameCheck):
    class_message = "All columns in ball data frame must have at least 1 truthy value."

    def critical_checks(self) -> List[Optional[AssertionError]]:
        ball = self.obj.ball
        return [
            self.check(condition, f'ball data frame column: {column_name} has truthy values')
            for column_name, condition in ball.any(axis=0).iteritems()
        ]



import unittest

from carball.analysis.analysis_manager import AnalysisManager
from carball.tests.utils import run_analysis_test_on_replay, get_specific_replays, get_specific_answers


class TeamTest(unittest.TestCase):

    def test_team_stat_creation(self):

        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            teams = proto_game.teams
            for team in teams:
                if team.is_orange:
                    continue
                self.assertGreater(team.stats.center_of_mass.average_max_distance_from_center, 0)
                self.assertGreater(team.stats.center_of_mass.average_distance_from_center, 0)
                self.assertGreater(team.stats.center_of_mass.positional_tendencies.time_behind_ball, 0)

        run_analysis_test_on_replay(test, get_specific_replays()["PASSES"])



import logging
import time
from typing import Dict

import pandas as pd
import json
import os

script_path = os.path.abspath(__file__)
with open(os.path.join(os.path.dirname(script_path), 'PROTOBUF_VERSION'), 'r') as f:
    PROTOBUF_VERSION = json.loads(f.read())

from ..analysis.cleaner.cleaner import clean_replay
from ..analysis.hit_detection.base_hit import BaseHit
from ..analysis.hit_detection.hit_analysis import SaltieHit
from ..analysis.saltie_game.metadata.ApiGame import ApiGame
from ..analysis.saltie_game.metadata.ApiMutators import ApiMutators
from ..analysis.saltie_game.metadata.ApiPlayer import ApiPlayer
from ..analysis.saltie_game.metadata.ApiTeam import ApiTeam
from ..analysis.saltie_game.saltie_game import SaltieGame
from ..analysis.stats.stats_manager import StatsManager
from ..analysis.utils.pandas_manager import PandasManager
from ..analysis.utils.proto_manager import ProtobufManager
from ..generated.api import game_pb2
from ..generated.api.player_pb2 import Player
from ..json_parser.game import Game

logger = logging.getLogger(__name__)


class AnalysisManager:
    id_creator = None
    timer = None

    def __init__(self, game: Game):
        self.game = game
        self.protobuf_game = game_pb2.Game()
        self.protobuf_game.version = PROTOBUF_VERSION
        self.id_creator = self.create_player_id_function(game)
        self.stats_manager = StatsManager()
        self.should_store_frames = False
        self.df_bytes = None

    def create_analysis(self):
        """
        Organizes all the different analysis that can occurs
        :return:
        """
        self.start_time()
        player_map = self.get_game_metadata(self.game, self.protobuf_game)
        self.log_time("creating metadata")
        data_frame = self.get_data_frames(self.game)
        self.log_time("getting frames")
        kickoff_frames, first_touch_frames = self.get_kickoff_frames(self.game, self.protobuf_game, data_frame)
        self.game.kickoff_frames = kickoff_frames
        self.log_time("getting kickoff")
        if self.can_do_full_analysis(first_touch_frames):
            self.perform_full_analysis(self.game, self.protobuf_game, player_map,
                                       data_frame, kickoff_frames, first_touch_frames)
        else:
            self.protobuf_game.game_metadata.is_invalid_analysis = True

        # log before we add the dataframes
        # logger.debug(self.protobuf_game)

        self.store_frames(data_frame)

    def perform_full_analysis(self, game: Game, proto_game: game_pb2.Game, player_map,
                              data_frame, kickoff_frames, first_touch_frames):
        """
        Performs the more in depth analysis on the game in addition to just metadata.
        :param game:
        :param proto_game:
        :param player_map:
        :param data_frame:
        :param kickoff_frames:
        :param first_touch_frames:
        :return:
        """
        self.get_game_time(proto_game, data_frame)
        clean_replay(game, data_frame, proto_game, player_map)
        self.calculate_hit_stats(game, proto_game, player_map, data_frame, kickoff_frames, first_touch_frames)
        self.log_time("calculating hits")
        self.get_stats(game, proto_game, player_map, data_frame)

    def get_game_metadata(self, game: Game, proto_game: game_pb2.Game) -> Dict[str, Player]:

        # create general metadata
        ApiGame.create_from_game(proto_game.game_metadata, game, self.id_creator)

        # create team metadata
        ApiTeam.create_teams_from_game(game, proto_game, self.id_creator)

        ApiGame.create_parties(proto_game.parties, game, self.id_creator)
        # create player metadata
        player_map = dict()
        for player in game.players:
            player_proto = proto_game.players.add()
            ApiPlayer.create_from_player(player_proto, player, self.id_creator)
            player_map[str(player.online_id)] = player_proto

        # create mutators
        ApiMutators.create_from_game(proto_game.mutators, game, self.id_creator)

        return player_map

    def get_game_time(self, protobuf_game: game_pb2.Game, data_frame: pd.DataFrame):
        protobuf_game.game_metadata.length = data_frame.game[data_frame.game.goal_number.notnull()].delta.sum()
        for player in protobuf_game.players:
            try:
                player.time_in_game = data_frame[
                    data_frame[player.name].pos_x.notnull() & data_frame.game.goal_number.notnull()].game.delta.sum()
                player.first_frame_in_game = data_frame[player.name].pos_x.first_valid_index()
            except:
                player.time_in_game = 0
        logger.info('created all times for players')

    def get_data_frames(self, game: Game):
        data_frame = SaltieGame.create_data_df(game)

        logger.info("Assigned goal_number in .data_frame")
        return data_frame

    def get_kickoff_frames(self, game: Game, proto_game: game_pb2.Game, data_frame: pd.DataFrame):
        kickoff_frames = SaltieGame.get_kickoff_frames(game)
        first_touch_frames = SaltieGame.get_first_touch_frames(game)

        for goal_number, goal in enumerate(game.goals):
            data_frame.loc[
            kickoff_frames[goal_number]: goal.frame_number, ('game', 'goal_number')
            ] = goal_number

        # Set goal_number of frames that are post last kickoff to -1 (ie non None)
        if len(kickoff_frames) > len(proto_game.game_metadata.goals):
            data_frame.loc[kickoff_frames[-1]:, ('game', 'goal_number')] = -1

        for index in range(min(len(kickoff_frames), len(first_touch_frames))):
            kickoff = proto_game.game_stats.kickoffs.add()
            kickoff.start_frame_number = kickoff_frames[index]
            kickoff.end_frame_number = first_touch_frames[index]

        return kickoff_frames, first_touch_frames

    def calculate_hit_stats(self, game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                            data_frame, kickoff_frames, first_touch_frames):
        logger.info("Looking for hits.")
        hits = BaseHit.get_hits_from_game(game, proto_game, self.id_creator, data_frame, first_touch_frames)
        logger.info("Found %s hits." % len(hits))

        SaltieHit.get_saltie_hits_from_game(proto_game, hits, player_map, data_frame, kickoff_frames)
        logger.info("Analysed hits.")

        # self.stats = get_stats(self)

    def get_stats(self, game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                  data_frame: pd.DataFrame):
        """
        Calculates all stats that are beyond the basic event creation.
        This is only active on valid goal frames.
        """
        goal_frames = data_frame.game.goal_number.notnull()
        self.stats_manager.get_stats(game, proto_game, player_map, data_frame[goal_frames])

    def store_frames(self, data_frame: pd.DataFrame):
        self.data_frame = data_frame
        self.df_bytes = PandasManager.safe_write_pandas_to_memory(data_frame)

    def write_proto_out_to_file(self, file):
        ProtobufManager.write_proto_out_to_file(file, self.protobuf_game)

    def write_pandas_out_to_file(self, file):
        if self.df_bytes is not None:
            file.write(self.df_bytes)
        elif not self.should_store_frames:
            logger.warning("pd DataFrames are not being stored anywhere")

    def get_protobuf_data(self) -> game_pb2.Game:
        """
        :return: The protobuf data created by the analysis
        """
        return self.protobuf_game

    def get_data_frame(self) -> pd.DataFrame:
        return self.data_frame

    def start_time(self):
        self.timer = time.time()
        logger.info("starting timer")

    def log_time(self, message=""):
        end = time.time()
        logger.info("Time taken for %s is %s milliseconds", message, (end - self.timer) * 1000)
        self.timer = end

    def create_player_id_function(self, game: Game):
        name_map = {player.name: player.online_id for player in game.players}

        def create_name(proto_player_id, name):
            proto_player_id.id = str(name_map[name])

        return create_name

    def can_do_full_analysis(self, first_touch_frames) -> bool:
        # Analyse only if 1v1 or 2v2 or 3v3
        team_sizes = []
        for team in self.game.teams:
            team_sizes.append(len(team.players))

        if len(team_sizes) == 0:
            logger.warning("Not doing full analysis. No teams found")
            return False
        if len(first_touch_frames) == 0:
            logger.warning("Not doing full analysis. No one touched the ball")
            return False
        # if any((team_size != team_sizes[0]) for team_size in team_sizes):
        #     logger.warning("Not doing full analysis. Not all team sizes are equal")
        #     return False

        return True



from typing import Dict

import pandas as pd

from carball.analysis.constants.playlist import get_team_size_from_game
from carball.analysis.stats.utils.pandas_utils import sum_deltas_by_player_name
from carball.generated.api.player_pb2 import Player

from carball.generated.api import game_pb2
from carball.json_parser.game import Game

from carball.generated.api.stats.team_stats_pb2 import TeamStats

from carball.analysis.stats.tendencies.team_tendencies import TeamTendencies


class RelativeTendencies(TeamTendencies):
    def calculate_team_stat(self, team_stat_list: Dict[int, TeamStats], game: Game, proto_game: game_pb2.Game,
                            player_map: Dict[str, Player], data_frame: pd.DataFrame):
        for team in proto_game.teams:
            team_size = get_team_size_from_game(proto_game)
            if len(team.player_ids) <= 2 or team_size <= 2:
                # center of mass does not matter for 1s games
                continue

            player_y = {player_map[player_id.id].name: data_frame[player_map[player_id.id].name].pos_y
                        for player_id in team.player_ids}

            player_y_data_frame = pd.concat(player_y, axis=1)
            last_person = player_y_data_frame.idxmin(axis=1).rename('last_person')
            first_person = player_y_data_frame.idxmax(axis=1).rename('first_person')

            position_distances_time = pd.concat([
                sum_deltas_by_player_name(data_frame, players_data_frame)
                for players_data_frame in [last_person, first_person]
            ], axis=1)

            position_distances_time.fillna(value=0, inplace=True)

            for player_id in team.player_ids:
                relative_position = player_map[player_id.id].stats.relative_positioning
                player_name = player_map[player_id.id].name
                try:
                    # flip first and last if you are orange
                    if team.is_orange:
                        relative_position.time_most_back_player = position_distances_time['first_person'][player_name]
                        relative_position.time_most_forward_player = position_distances_time['last_person'][player_name]
                    else:
                        relative_position.time_most_back_player = position_distances_time['last_person'][player_name]
                        relative_position.time_most_forward_player = position_distances_time['first_person'][player_name]

                    total_time = relative_position.time_most_back_player + relative_position.time_most_forward_player
                    relative_position.time_between_players = player_map[player_id.id].time_in_game - total_time
                except (AttributeError, KeyError) as e:
                    pass



import numpy as np

R = 91.25
Y = 2.0
mu = 0.285
C_R = 0.6
A = 0.0003

GROUND_NORMAL = np.array([0, 0, 1])


def bounce(state, normal=GROUND_NORMAL):
    # pos = state.loc[:, ['x', 'y', 'z']].values
    # vel = state.loc[:, ['vx', 'vy', 'vz']].values
    # ang_vel = state.loc[:, ['rotvx', 'rotvy', 'rotvz']].values
    vel, ang_vel = state
    normal = normal.astype(float)
    ang_vel = ang_vel.astype(float)
    v_perp = np.dot(vel, normal) * normal
    v_para = vel - v_perp
    v_spin = R * np.cross(normal, ang_vel)
    s = v_para + v_spin

    ratio = np.sqrt(v_perp.dot(v_perp)) / np.sqrt(s.dot(s))

    delta_v_perp = - (1.0 + C_R) * v_perp
    delta_v_para = - min(1.0, Y * ratio) * mu * s

    new_state = (vel + delta_v_perp + delta_v_para,
                 ang_vel + A * R * np.cross(delta_v_para.astype(float), normal))
    return new_state



import logging
import time
from typing import Dict, List

import pandas as pd

from ...analysis.stats.stats_list import StatsList
from ...generated.api import game_pb2
from ...generated.api.player_pb2 import Player
from ...generated.api.stats.player_stats_pb2 import PlayerStats
from ...generated.api.stats.team_stats_pb2 import TeamStats
from ...generated.api.team_pb2 import Team
from ...json_parser.game import Game

logger = logging.getLogger(__name__)


def start_time():
    return time.time()


def end_time(start):
    return (time.time() - start) * 1000


class StatsManager:

    def get_stats(self, game: Game, proto_game: game_pb2.Game,
                  player_map: Dict[str, Player], data_frame: pd.DataFrame):
        """
        Calculates all basic stats.
        The stats are always calculated in this order:
            Player, Team, Game, Hit
        """
        self.calculate_player_stats(game, proto_game, player_map, data_frame)
        self.calculate_team_stats(game, proto_game, proto_game.teams, player_map, data_frame)
        self.calculate_game_stats(game, proto_game, player_map, data_frame)
        self.calculate_hit_stats(game, proto_game, player_map, data_frame)

    @staticmethod
    def calculate_player_stats(game: Game, proto_game: game_pb2.Game,
                               player_map: Dict[str, Player], data_frame: pd.DataFrame):
        stats_proto: Dict[str, PlayerStats] = {
            key: player.stats
            for key, player in player_map.items()
        }
        for stat_function in StatsList.get_player_stats():
            time = start_time()
            logger.debug("Building player stat: %s", type(stat_function).__name__)
            stat_function.calculate_player_stat(stats_proto, game, proto_game, player_map, data_frame)
            logger.debug("Built in [%d] milliseconds", end_time(time))

    @staticmethod
    def calculate_team_stats(game: Game, proto_game: game_pb2.Game, teams: List[Team],
                             player_map: Dict[str, Player], data_frame: pd.DataFrame):
        stats_proto: Dict[int, TeamStats] = {
            int(team.is_orange): team.stats
            for team in teams
        }
        for stat_function in StatsList.get_team_stats():
            time = start_time()
            logger.debug("Building team stat: %s", type(stat_function).__name__)
            stat_function.calculate_team_stat(stats_proto, game, proto_game, player_map, data_frame)
            logger.debug("Built in [%d] milliseconds", end_time(time))

    @staticmethod
    def calculate_game_stats(game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                             data_frame: pd.DataFrame):
        for stat_function in StatsList.get_general_stats():
            time = start_time()
            logger.debug("Building game stat: %s", type(stat_function).__name__)
            stat_function.calculate_stat(proto_game.game_stats, game, proto_game, player_map, data_frame)
            logger.debug("Built in [%d] milliseconds", end_time(time))

    @staticmethod
    def calculate_hit_stats(game: Game, proto_game: game_pb2.Game,
                            player_map: Dict[str, Player], data_frame: pd.DataFrame):
        hit_stats = StatsList.get_hit_stats()
        for hit_stat in hit_stats:
            logger.debug("Building hit stat: %s", type(hit_stat).__name__)
            hit_stat.initialize_hit_stat(game, player_map, data_frame)
        hits = proto_game.game_stats.hits
        for hit_index in range(len(hits) - 1):
            current_hit = hits[hit_index]
            if current_hit.HasField("next_hit_frame_number"):
                next_hit = hits[hit_index + 1]
                for hit_stat in hit_stats:
                    hit_stat.calculate_next_hit_stat(game, proto_game, current_hit, next_hit, player_map, hit_index)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/stats/team_stats.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ...api.stats import stats_pb2
from ...api.stats import per_possession_stats_pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/stats/team_stats.proto',
  package='api.stats',
  serialized_pb=_b('\n\x1a\x61pi/stats/team_stats.proto\x12\tapi.stats\x1a\x15\x61pi/stats/stats.proto\x1a$api/stats/per_possession_stats.proto\"\xcc\x01\n\x0c\x43\x65nterOfMass\x12>\n\x15positional_tendencies\x18\x01 \x01(\x0b\x32\x1f.api.stats.PositionalTendencies\x12$\n\x1c\x61verage_distance_from_center\x18\x02 \x01(\x02\x12(\n average_max_distance_from_center\x18\x03 \x01(\x02\x12\x14\n\x0ctime_clumped\x18\x04 \x01(\x02\x12\x16\n\x0etime_boondocks\x18\x05 \x01(\x02\"\xce\x01\n\tTeamStats\x12)\n\npossession\x18\x01 \x01(\x0b\x32\x15.api.stats.Possession\x12(\n\nhit_counts\x18\x02 \x01(\x0b\x32\x14.api.stats.HitCounts\x12/\n\x0e\x63\x65nter_of_mass\x18\x03 \x01(\x0b\x32\x17.api.stats.CenterOfMass\x12;\n\x14per_possession_stats\x18\x04 \x01(\x0b\x32\x1d.api.stats.PerPossessionStats')
  ,
  dependencies=[stats_pb2.DESCRIPTOR,per_possession_stats_pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_CENTEROFMASS = _descriptor.Descriptor(
  name='CenterOfMass',
  full_name='api.stats.CenterOfMass',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='positional_tendencies', full_name='api.stats.CenterOfMass.positional_tendencies', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='average_distance_from_center', full_name='api.stats.CenterOfMass.average_distance_from_center', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='average_max_distance_from_center', full_name='api.stats.CenterOfMass.average_max_distance_from_center', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_clumped', full_name='api.stats.CenterOfMass.time_clumped', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_boondocks', full_name='api.stats.CenterOfMass.time_boondocks', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=103,
  serialized_end=307,
)


_TEAMSTATS = _descriptor.Descriptor(
  name='TeamStats',
  full_name='api.stats.TeamStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='possession', full_name='api.stats.TeamStats.possession', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='hit_counts', full_name='api.stats.TeamStats.hit_counts', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='center_of_mass', full_name='api.stats.TeamStats.center_of_mass', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='per_possession_stats', full_name='api.stats.TeamStats.per_possession_stats', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=310,
  serialized_end=516,
)

_CENTEROFMASS.fields_by_name['positional_tendencies'].message_type = stats_pb2._POSITIONALTENDENCIES
_TEAMSTATS.fields_by_name['possession'].message_type = stats_pb2._POSSESSION
_TEAMSTATS.fields_by_name['hit_counts'].message_type = stats_pb2._HITCOUNTS
_TEAMSTATS.fields_by_name['center_of_mass'].message_type = _CENTEROFMASS
_TEAMSTATS.fields_by_name['per_possession_stats'].message_type = per_possession_stats_pb2._PERPOSSESSIONSTATS
DESCRIPTOR.message_types_by_name['CenterOfMass'] = _CENTEROFMASS
DESCRIPTOR.message_types_by_name['TeamStats'] = _TEAMSTATS

CenterOfMass = _reflection.GeneratedProtocolMessageType('CenterOfMass', (_message.Message,), dict(
  DESCRIPTOR = _CENTEROFMASS,
  __module__ = 'api.stats.team_stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.CenterOfMass)
  ))
_sym_db.RegisterMessage(CenterOfMass)

TeamStats = _reflection.GeneratedProtocolMessageType('TeamStats', (_message.Message,), dict(
  DESCRIPTOR = _TEAMSTATS,
  __module__ = 'api.stats.team_stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.TeamStats)
  ))
_sym_db.RegisterMessage(TeamStats)


# @@protoc_insertion_point(module_scope)



import unittest

import numpy as np
from ..analysis.utils.split_location import LocationSplitManager, STANDARD_MIN_VALUES, STANDARD_MAX_VALUES


class DBTest(unittest.TestCase):
    def __init__(self, *args):
        super().__init__(*args)
        self.split_manager = None

    def setUp(self):
        self.split_manager = LocationSplitManager(3, np.array([7, 7, 5]), STANDARD_MIN_VALUES, STANDARD_MAX_VALUES)

    def test_replay_attrs(self):
        result = self.split_manager.create_boxes(np.array([0, 0, 0]))
        answer = [np.array([3., 3., 0.]), np.array([3., 3., 0.]), np.array([3., 3., 0.])]
        for i in range(len(result)):
            np.testing.assert_equal(answer[i], result[i])
        print(result)


if __name__ == '__main__':
    unittest.main()






# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/stats/game_stats.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ...api.stats import events_pb2
from ...api.stats import ball_stats_pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/stats/game_stats.proto',
  package='api.stats',
  serialized_pb=_b('\n\x1a\x61pi/stats/game_stats.proto\x12\tapi.stats\x1a\x16\x61pi/stats/events.proto\x1a\x1a\x61pi/stats/ball_stats.proto\"\xba\x01\n\tGameStats\x12\x1c\n\x04hits\x18\x01 \x03(\x0b\x32\x0e.api.stats.Hit\x12\x1f\n\x17neutral_possession_time\x18\x02 \x01(\x02\x12\x1e\n\x05\x62umps\x18\x03 \x03(\x0b\x32\x0f.api.stats.Bump\x12(\n\nball_stats\x18\x04 \x01(\x0b\x32\x14.api.stats.BallStats\x12$\n\x08kickoffs\x18\x05 \x03(\x0b\x32\x12.api.stats.Kickoff')
  ,
  dependencies=[events_pb2.DESCRIPTOR,ball_stats_pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_GAMESTATS = _descriptor.Descriptor(
  name='GameStats',
  full_name='api.stats.GameStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='hits', full_name='api.stats.GameStats.hits', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='neutral_possession_time', full_name='api.stats.GameStats.neutral_possession_time', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='bumps', full_name='api.stats.GameStats.bumps', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ball_stats', full_name='api.stats.GameStats.ball_stats', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='kickoffs', full_name='api.stats.GameStats.kickoffs', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=94,
  serialized_end=280,
)

_GAMESTATS.fields_by_name['hits'].message_type = events_pb2._HIT
_GAMESTATS.fields_by_name['bumps'].message_type = events_pb2._BUMP
_GAMESTATS.fields_by_name['ball_stats'].message_type = ball_stats_pb2._BALLSTATS
_GAMESTATS.fields_by_name['kickoffs'].message_type = events_pb2._KICKOFF
DESCRIPTOR.message_types_by_name['GameStats'] = _GAMESTATS

GameStats = _reflection.GeneratedProtocolMessageType('GameStats', (_message.Message,), dict(
  DESCRIPTOR = _GAMESTATS,
  __module__ = 'api.stats.game_stats_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.GameStats)
  ))
_sym_db.RegisterMessage(GameStats)


# @@protoc_insertion_point(module_scope)



import os
from urllib import request
import json
from distutils.version import StrictVersion

from carball.rattletrap.rattletrap_utils import get_rattletrap_binaries, get_rattletrap_path


def update_rattletrap():
    path = get_rattletrap_path()
    print('updating rattletrap in path', path)
    files = os.listdir(path)
    print('files in path', files)

    cur_ver = '0.0.0'
    binaries = get_rattletrap_binaries(path)
    print('existing found', binaries)

    response = request.urlopen('https://api.github.com/repos/tfausak/rattletrap/releases/latest')

    js = json.loads(response.read())

    if len(binaries) > 0:
        cur_ver = binaries[0].split('-')[1]
    update = StrictVersion(js['name']) > StrictVersion(cur_ver)
    print (f'GitHub version: {js["name"]}\n'
           f'Current version: {cur_ver}\n'
           f'Update? {update}')
    if update:
        for file in binaries:
            os.remove(file)

        for asset in js['assets']:
            print('Downloading {}'.format(asset['name']))
            new_file = os.path.join(path, asset['name'])
            request.urlretrieve(asset['browser_download_url'], filename=new_file)
            print('making file executable', new_file)
            os.chmod(new_file, 0o777)


if __name__ == "__main__":
    update_rattletrap()



import unittest

from carball.analysis.analysis_manager import AnalysisManager

from carball.tests.utils import run_analysis_test_on_replay, get_specific_replays, get_specific_answers, get_raw_replays


class HitTest(unittest.TestCase):

    def test_num_hits_detected(self):
        def test(analysis: AnalysisManager, answer):
            proto_game = analysis.get_protobuf_data()
            hits = proto_game.game_stats.hits
            self.assertEqual(len(hits), answer)
            print(analysis)

        run_analysis_test_on_replay(test, get_specific_replays()["HITS"], get_specific_answers()["HITS"])

    def test_num_kickoff_hits_detected(self):
        def test(analysis: AnalysisManager):
            proto_game = analysis.get_protobuf_data()
            hits = proto_game.game_stats.hits
            kickoff_counter = 0
            for hit in hits:
                if hit.is_kickoff:
                    kickoff_counter += 1
            print(analysis)

            self.assertEqual(kickoff_counter, 3)

        run_analysis_test_on_replay(test, get_raw_replays()["3_KICKOFFS"])

    def test_num_shots_detected(self):
        def test(analysis: AnalysisManager, answer):
            proto_game = analysis.get_protobuf_data()
            hits = proto_game.game_stats.hits
            shot_counter = 0
            for hit in hits:
                if hit.shot:
                    shot_counter += 1

            self.assertEqual(shot_counter, answer)

        run_analysis_test_on_replay(test, get_specific_replays()["SHOTS"], get_specific_answers()["SHOTS"])

    def test_num_passes_detected(self):
        def test(analysis: AnalysisManager, answer):
            proto_game = analysis.get_protobuf_data()
            hits = proto_game.game_stats.hits
            pass_counter = 0
            for hit in hits:
                if hit.pass_:
                    pass_counter += 1

            self.assertEqual(pass_counter, answer)

        run_analysis_test_on_replay(test, get_specific_replays()["PASSES"], get_specific_answers()["PASSES"])

    def test_num_saves_detected(self):
        def test(analysis: AnalysisManager, answer):
            proto_game = analysis.get_protobuf_data()
            hits = proto_game.game_stats.hits
            counter = 0
            for hit in hits:
                if hit.save:
                    counter += 1

            self.assertEqual(counter, answer)

        run_analysis_test_on_replay(test, get_specific_replays()["SAVES"], get_specific_answers()["SAVES"])

    def test_num_aerials_detected(self):
        def test(analysis: AnalysisManager, answer):
            proto_game = analysis.get_protobuf_data()
            hits = proto_game.game_stats.hits
            aerial_counter = 0
            for hit in hits:
                if hit.aerial:
                    aerial_counter += 1

            self.assertEqual(aerial_counter, answer)

        run_analysis_test_on_replay(test, get_specific_replays()["AERIALS"], get_specific_answers()["AERIALS"])


if __name__ == '__main__':
    unittest.main()



import logging

logger = logging.getLogger(__name__)


class GameInfo:
    def __init__(self):
        self.server_id = None
        self.server_name = None
        self.match_guid = None
        self.game_mode = None
        self.mutator_index = None

    def parse_game_info_actor(self, actor_data: dict):
        """
        Parses game actor
        :param actor_data: game replication info
        :return: self
        """        
        # There is no GameServerID if you play alone
        self.server_id = ''
        if 'ProjectX.GRI_X:GameServerID' in actor_data:
            self.server_id = str(actor_data['ProjectX.GRI_X:GameServerID']['q_word'])
        self.server_name = actor_data['Engine.GameReplicationInfo:ServerName']
        
        # A custom lobby doesn't have a MatchGUID
        self.match_guid = actor_data.get('ProjectX.GRI_X:MatchGUID', '')
        self.playlist = actor_data['ProjectX.GRI_X:ReplicatedGamePlaylist']
        self.mutator_index = actor_data.get('ProjectX.GRI_X:ReplicatedGameMutatorIndex', 0)

        logger.info('Created game info from actor')
        return self




import logging
import time
from typing import List, Dict, Callable

import numpy as np
import pandas as pd

from carball.analysis.constants.basic_math import position_column_names, get_player_ball_displacements, \
    positional_columns
from .hitbox.hitbox import Hitbox
from ...generated.api import game_pb2
from ...generated.api.stats.events_pb2 import Hit
from ...json_parser.game import Game

logger = logging.getLogger(__name__)

MIN_DRIBBLE_FRAME_DISTANCE = 10


class BaseHit:

    @staticmethod
    def get_hits_from_game(game: Game, proto_game: game_pb2, id_creation: Callable,
                           data_frame: pd.DataFrame, first_touch_frames: pd.Series) -> Dict[int, Hit]:

        start_time = time.time()

        team_dict = {}

        for team in game.teams:
            team_dict[team.is_orange] = team

        hit_frame_numbers = BaseHit.get_hit_frame_numbers_by_ball_ang_vel(data_frame)

        # add kickoff hits
        for hit_frame in first_touch_frames:
            if hit_frame not in hit_frame_numbers:
                hit_frame_numbers.append(hit_frame)

        hit_frame_numbers.sort()

        hit_creation_time = time.time()
        logger.info('time to get get frame_numbers: %s', (hit_creation_time - start_time) * 1000)

        hit_frames = data_frame.loc[hit_frame_numbers, (slice(None), positional_columns)]
        player_displacements = {player.name: get_player_ball_displacements(hit_frames, player.name)
                                for player in game.players}

        # player_distances = {player_name: get_distance_from_displacements(data_frame).rename(player_name)
        #                     for player_name, data_frame in player_displacements.items()}

        # player_distances_data_frame = pd.concat(player_distances, axis=1)

        rotation_matrices = {player.name: get_rotation_matrices(hit_frames, player.name) for player in game.players}

        local_displacements: Dict[str, pd.DataFrame] = {
            player.name: get_local_displacement(player_displacements[player.name],
                                                rotation_matrices[player.name])
            for player in game.players
        }

        player_hitboxes = get_player_hitboxes(game)
        collision_distances = [
            get_collision_distances(local_displacements[player.name], player_hitboxes[player.name]).rename(player.name)
            for player in game.players
        ]
        collision_distances_data_frame = pd.concat(collision_distances, axis=1)
        # TODO: Fix when no players detected. See issue #115

        player_name_to_team: Dict[str, int] = {player.name: int(player.team.is_orange) for player in game.players}
        columns = [(player_name_to_team[player_name], player_name)
                   for player_name in collision_distances_data_frame.columns]
        collision_distances_data_frame.columns = pd.MultiIndex.from_tuples(columns)

        collision_distances_data_frame['closest_player', 'name'] = None
        collision_distances_data_frame['closest_player', 'distance'] = None
        for hit_team_no in [0, 1]:
            try:
                collision_distances_for_team = collision_distances_data_frame[
                    hit_team_no].loc[data_frame.ball['hit_team_no'] == hit_team_no]

                close_collision_distances_for_team = collision_distances_for_team[
                    (collision_distances_for_team < 300).any(axis=1)
                ]

                collision_distances_data_frame['closest_player', 'distance'].fillna(
                    close_collision_distances_for_team.min(axis=1),
                    inplace=True
                )
                collision_distances_data_frame['closest_player', 'name'].fillna(
                    close_collision_distances_for_team.idxmin(axis=1),
                    inplace=True
                )
            except KeyError as e:
                if e.args[0] == hit_team_no:
                    logger.warning("Team %s did not hit the ball", str(hit_team_no))
                else:
                    raise e

        all_hits = {}
        hits_data = collision_distances_data_frame['closest_player'].dropna()
        if len(hits_data) > 1:
            hit_frames_to_keep = BaseHit.filter_out_duplicate_hits(hits_data)
            hits_data = hits_data.loc[hit_frames_to_keep]

        for row in hits_data.itertuples():
            frame_number, player_name, collision_distance = row.Index, row.name, row.distance
            hit = proto_game.game_stats.hits.add()
            hit.frame_number = frame_number
            goal_number = data_frame.at[frame_number, ('game', 'goal_number')]
            if not np.isnan(goal_number):
                hit.goal_number = int(goal_number)
            id_creation(hit.player_id, player_name)
            hit.collision_distance = collision_distance
            ball_position = data_frame.ball.loc[frame_number, position_column_names]
            hit.ball_data.pos_x = float(ball_position['pos_x'])
            hit.ball_data.pos_y = float(ball_position['pos_y'])
            hit.ball_data.pos_z = float(ball_position['pos_z'])
            hit.is_kickoff = hit.frame_number in first_touch_frames
            all_hits[frame_number] = hit

        time_diff = time.time() - hit_creation_time
        logger.info('ball hit creation time: %s', time_diff * 1000)
        return all_hits

    @staticmethod
    def filter_out_duplicate_hits(hits_data):
        """
        Filters out duplicate hits by finding the min distance in a set of frames by the same person
        That is less time than the MIN_DRIBBLE_FRAME_DISTANCE
        :param hits_data: Suspected hits
        :return: A reduced list that has removed duplicate hits.
        """
        hit_frames_to_keep = []
        shifted_hits = hits_data.shift(1)
        different_hits = list(hits_data[hits_data['name'] != shifted_hits['name']].index)
        if different_hits[-1] != int(hits_data.index[-1]):
            different_hits += [int(hits_data.index[-1])]

        if different_hits[0] != int(hits_data.index[0]):
            different_hits = [int(hits_data.index[0])] + different_hits

        for index in range(len(different_hits) - 1):
            start_row_num = different_hits[index]
            end_row_num = different_hits[index + 1]
            hit_rows = hits_data.loc[start_row_num:end_row_num]

            # ignore last frame in hit row for all but very last group
            if index + 1 < len(different_hits) - 1:
                hit_rows = hit_rows[:-1]

            if len(hit_rows) == 1:
                hit_frames_to_keep.append(start_row_num)
            else:
                min_distance = None
                min_frame_value = None
                starting_frame_number = 0
                last_added_frame_number = 0
                old_name = None
                # go through a single person hit
                for row in hit_rows.itertuples():
                    frame_number, player_name, collision_distance = row.Index, row.name, row.distance

                    if min_frame_value is None:
                        min_frame_value = frame_number
                        min_distance = collision_distance
                        starting_frame_number = frame_number
                        old_name = row.name
                        continue

                    if frame_number - starting_frame_number > MIN_DRIBBLE_FRAME_DISTANCE or row.name != old_name:
                        # Same person but a new hit reset distance checks
                        hit_frames_to_keep.append(min_frame_value)
                        last_added_frame_number = min_frame_value

                        # reset for next shot
                        min_frame_value = frame_number
                        min_distance = collision_distance
                        starting_frame_number = frame_number
                        old_name = row.name
                        continue

                    if collision_distance < min_distance:
                        min_frame_value = frame_number
                        min_distance = collision_distance

                if last_added_frame_number != min_frame_value:
                    hit_frames_to_keep.append(min_frame_value)

            start_row_num += 1
        return hit_frames_to_keep

    @staticmethod
    def get_hit_frame_numbers_by_ball_ang_vel(data_frame: pd.DataFrame) -> List[int]:
        ball_ang_vels = data_frame.ball.loc[:, ['ang_vel_x', 'ang_vel_y', 'ang_vel_z']]
        diff_series = ball_ang_vels.diff().any(axis=1)
        indices = diff_series.index[diff_series].tolist()
        return indices

    @staticmethod
    def get_ball_data(data_frame: pd.DataFrame, hit: Hit):
        return data_frame.ball.loc[hit.frame_number, :]


def get_rotation_matrices(data_frame: pd.DataFrame, player_name: str) -> pd.Series:
    pitch = data_frame[player_name, 'rot_x']
    yaw = data_frame[player_name, 'rot_y']
    roll = data_frame[player_name, 'rot_z']

    cos_roll = np.cos(roll).rename('cos_roll')
    sin_roll = np.sin(roll).rename('sin_roll')
    cos_pitch = np.cos(pitch).rename('cos_pitch')
    sin_pitch = np.sin(pitch).rename('sin_pitch')
    cos_yaw = np.cos(yaw).rename('cos_yaw')
    sin_yaw = np.sin(yaw).rename('sin_yaw')

    components: pd.DataFrame = pd.concat([cos_roll, sin_roll, cos_pitch, sin_pitch, cos_yaw, sin_yaw], axis=1)

    rotation_matrix = components.apply(get_rotation_matrix_from_row, axis=1, result_type='reduce')
    return rotation_matrix


def get_rotation_matrix_from_row(components: pd.Series) -> np.array:
    cos_roll, sin_roll, cos_pitch, sin_pitch, cos_yaw, sin_yaw = components.values
    rotation_matrix = np.array(
        [[cos_pitch * cos_yaw, cos_yaw * sin_pitch * sin_roll - cos_roll * sin_yaw,
          -cos_roll * cos_yaw * sin_pitch - sin_roll * sin_yaw],
         [cos_pitch * sin_yaw, sin_yaw * sin_pitch * sin_roll + cos_roll * cos_yaw,
          -cos_roll * sin_yaw * sin_pitch + sin_roll * cos_yaw],
         [sin_pitch, -cos_pitch * sin_roll, cos_pitch * cos_roll]])
    return rotation_matrix


def get_local_displacement(displacement: pd.DataFrame, rotation_matrices: pd.Series) -> pd.DataFrame:
    displacement_vectors = np.expand_dims(displacement[position_column_names].values, 2)
    inverse_rotation_matrices: pd.Series = np.transpose(rotation_matrices)
    inverse_rotation_array = np.stack(inverse_rotation_matrices.values)
    local_displacement = np.matmul(inverse_rotation_array, displacement_vectors)
    displacement_data_frame = pd.DataFrame(data=np.squeeze(local_displacement, 2),
                                           index=displacement.index,
                                           columns=position_column_names)
    return displacement_data_frame


def get_player_hitboxes(game: Game) -> Dict[str, Hitbox]:
    player_hitboxes = {}
    for player in game.players:
        car_item_id = player.loadout[0]['car'] if len(player.loadout) == 1 else player.loadout[player.is_orange]['car']
        player_hitboxes[player.name] = Hitbox(car_item_id)
    return player_hitboxes


def get_collision_distances(local_ball_displacement: pd.DataFrame, player_hitbox: Hitbox) -> pd.Series:
    def get_distance_function_for_player(displacement: pd.Series):
        return player_hitbox.get_collision_distance(displacement.values)

    collision_distances = local_ball_displacement.apply(get_distance_function_for_player, axis=1, result_type='reduce')
    return collision_distances



import numpy as np
import pandas as pd

from .bounce import bounce

BALL_RADIUS = 91.25
SIDE_WALL_DISTANCE = 4096
BACK_WALL_DISTANCE = 5140
CEILING_DISTANCE = 2044
CORNER_WALL_DISTANCE = 8000
GOAL_X = 892.75
GOAL_X = 950
GOAL_Z = 640
GOAL_Z = 750
GOAL_Z = 900
x = 590
# CURVE_RADIUS_1, CURVE_RADIUS_2, CURVE_RADIUS_3 = 520, 260, 190  # ramp radii
CURVE_RADIUS_1, CURVE_RADIUS_2, CURVE_RADIUS_3 = x, x / 2, 175  # ramp radii

CURVE_X_1 = SIDE_WALL_DISTANCE - CURVE_RADIUS_1
CURVE_X_2 = SIDE_WALL_DISTANCE - CURVE_RADIUS_2
CURVE_X_3 = SIDE_WALL_DISTANCE - CURVE_RADIUS_3
CURVE_Y_1 = BACK_WALL_DISTANCE - CURVE_RADIUS_1
# CURVE_Y_2 = BACK_WALL_DISTANCE - CURVE_RADIUS_2
CURVE_Y_3 = BACK_WALL_DISTANCE - CURVE_RADIUS_3
CURVE_Z_1 = CEILING_DISTANCE - CURVE_RADIUS_1
CURVE_Z_2 = CURVE_RADIUS_2
CURVE_Z_3 = CURVE_RADIUS_3

SIMULATION_FPS = 120
SIMULATION_SECONDS = 3


class BallSimulator:
    gravity = 650  # uu/s2
    air_resistance = 0.0305  # % loss per second
    ball_max_speed = 6000
    ball_max_rotation_speed = 6

    def __init__(self, ball_data, is_orange):
        self.is_orange = is_orange
        self.ball_data = ball_data
        self.sim_vars = {'position': ball_data[['pos_x', 'pos_y', 'pos_z']].values,
                         'velocity': ball_data[['vel_x', 'vel_y', 'vel_z']].values / 10,
                         'rotation': ball_data[['rot_x', 'rot_y', 'rot_z']].values,
                         'ang_vel': ball_data[['ang_vel_x', 'ang_vel_y', 'ang_vel_z']].values / 1000,
                         }
        self.is_shot = False

    def get_is_shot(self):
        """
        Simulates ball and returns is_shot. is_shot is updated in predict_ball_positions
        :return:
        """
        self.sim_data = self.predict_ball_positions(t=SIMULATION_SECONDS)
        return self.is_shot

    def predict_ball_positions(self, t):
        """
        Returns an array of time and position and velocity up to time=t.
        :param t: Number of seconds to simulate
        :return:
        """
        starting_x_v = np.concatenate((self.sim_vars['position'], self.sim_vars['velocity']))
        sim_data = self.simulate_time(0, t, 1 / SIMULATION_FPS, self.step_dt,
                                      starting_x_v)
        return sim_data

    def simulate_time(self, start_time, end_time, time_step, step_func, starting_values, break_if_goal=True,
                      get_sim_data=False):
        t_s = []
        x_vs = []
        av_s = []

        simulated_time = start_time
        latest_x_v = starting_values
        while simulated_time < end_time:
            # record values at current time
            t_s.append(simulated_time)
            x_vs.append(latest_x_v)
            av_s.append(self.sim_vars['ang_vel'])

            # move by dt
            derivatives = step_func(latest_x_v, simulated_time)
            latest_x_v[3:] = derivatives[:3]
            latest_x_v = latest_x_v + derivatives * time_step
            simulated_time += time_step

            # CHECK IF BALL IN GOAL
            if (self.is_orange and latest_x_v[1] < -BACK_WALL_DISTANCE) or (
                    (not self.is_orange) and latest_x_v[1] > BACK_WALL_DISTANCE):
                self.is_shot = True
                if break_if_goal:
                    break

        if not get_sim_data:
            return

        t_s = np.array(t_s)
        x_vs = np.array(x_vs)
        av_s = np.array(av_s)
        sim_data = pd.DataFrame(
            data=np.column_stack((t_s, x_vs, av_s)),
            columns=['t', 'x', 'y', 'z', 'vx', 'vy', 'vz', 'rotvx', 'rotvy', 'rotvz']
        )
        return sim_data

    def step_dt(self, x_v, t):
        x = x_v[:3]
        v = x_v[3:]
        # calculate collisions
        collided = False
        # ramps
        # bottom y axis
        if x[1] > CURVE_Y_3 and x[2] < CURVE_Z_3 and abs(x[0]) > GOAL_X and \
                (abs(x[1]) - CURVE_Y_3) ** 2 + (x[2] - CURVE_Z_3) ** 2 > (CURVE_RADIUS_3 - BALL_RADIUS) ** 2:
            surface_vector = np.array([0, CURVE_Y_3 - x[1], CURVE_Z_3 - x[2]])
            normal_vector = surface_vector / np.sqrt(surface_vector.dot(surface_vector))
            # print(t, x, normal_vector, surface_vector)
            # print(t, 'y+ bottom')
            if self.check_if_ball_leaving(x_v, normal_vector):
                collided = True
        elif x[1] < -CURVE_Y_3 and x[2] < CURVE_Z_3 and abs(x[0]) > GOAL_X and \
                (abs(x[1]) - CURVE_Y_3) ** 2 + (x[2] - CURVE_Z_3) ** 2 > (CURVE_RADIUS_3 - BALL_RADIUS) ** 2:
            surface_vector = np.array([0, CURVE_Y_3 - x[1], CURVE_Z_3 - x[2]])
            normal_vector = surface_vector / np.sqrt(surface_vector.dot(surface_vector))
            # print(t, x, normal_vector, surface_vector)
            # print(t, 'y- bottom')
            if self.check_if_ball_leaving(x_v, normal_vector):
                collided = True
        # bottom x axis
        if x[0] > CURVE_X_2 and x[2] < CURVE_Z_2 and \
                (abs(x[0]) - CURVE_X_2) ** 2 + (x[2] - CURVE_Z_2) ** 2 > (CURVE_RADIUS_2 - BALL_RADIUS) ** 2:
            surface_vector = np.array([CURVE_X_2 - x[0], 0, CURVE_Z_2 - x[2]])
            normal_vector = surface_vector / np.sqrt(surface_vector.dot(surface_vector))
            # print(t, x, normal_vector, surface_vector)
            # print(t, 'x+ bottom')
            if self.check_if_ball_leaving(x_v, normal_vector):
                collided = True
        elif x[0] < -CURVE_X_2 and x[2] < CURVE_Z_2 and \
                (abs(x[0]) - CURVE_X_2) ** 2 + (x[2] - CURVE_Z_2) ** 2 > (CURVE_RADIUS_2 - BALL_RADIUS) ** 2:
            surface_vector = np.array([CURVE_X_2 - x[0], 0, CURVE_Z_2 - x[2]])
            normal_vector = surface_vector / np.sqrt(surface_vector.dot(surface_vector))
            # print(t, x, normal_vector, surface_vector)
            # print(t, 'x- bottom')
            if self.check_if_ball_leaving(x_v, normal_vector):
                collided = True
        # top y axis
        if x[1] > CURVE_Y_1 and x[2] > CURVE_Z_1 and abs(x[0]) > GOAL_X and \
                (abs(x[1]) - CURVE_Y_1) ** 2 + (x[2] - CURVE_Z_1) ** 2 > (CURVE_RADIUS_1 - BALL_RADIUS) ** 2:
            surface_vector = np.array([0, CURVE_Y_1 - x[1], CURVE_Z_1 - x[2]])
            normal_vector = surface_vector / np.sqrt(surface_vector.dot(surface_vector))
            # print(t, x, normal_vector, surface_vector)
            # print(t, 'y+ top')
            if self.check_if_ball_leaving(x_v, normal_vector):
                collided = True
        elif x[1] < -CURVE_Y_1 and x[2] > CURVE_Z_1 and abs(x[0]) > GOAL_X and \
                (abs(x[1]) - CURVE_Y_1) ** 2 + (x[2] - CURVE_Z_1) ** 2 > (CURVE_RADIUS_1 - BALL_RADIUS) ** 2:
            surface_vector = np.array([0, CURVE_Y_1 - x[1], CURVE_Z_1 - x[2]])
            normal_vector = surface_vector / np.sqrt(surface_vector.dot(surface_vector))
            # print(t, x, normal_vector, surface_vector)
            # print(t, 'y- top')
            if self.check_if_ball_leaving(x_v, normal_vector):
                collided = True
        # top x axis
        if x[0] > CURVE_X_1 and x[2] > CURVE_Z_1 and \
                (abs(x[0]) - CURVE_X_1) ** 2 + (x[2] - CURVE_Z_1) ** 2 > (CURVE_RADIUS_1 - BALL_RADIUS) ** 2:
            surface_vector = np.array([CURVE_X_1 - x[0], 0, CURVE_Z_1 - x[2]])
            normal_vector = surface_vector / np.sqrt(surface_vector.dot(surface_vector))
            # print(t, x, normal_vector, surface_vector)
            # print(t, 'x+ top')
            if self.check_if_ball_leaving(x_v, normal_vector):
                collided = True
        elif x[0] < -CURVE_X_1 and x[2] > CURVE_Z_1 and \
                (abs(x[0]) - CURVE_X_1) ** 2 + (x[2] - CURVE_Z_1) ** 2 > (CURVE_RADIUS_1 - BALL_RADIUS) ** 2:
            surface_vector = np.array([CURVE_X_2 - x[0], 0, CURVE_Z_1 - x[2]])
            normal_vector = surface_vector / np.sqrt(surface_vector.dot(surface_vector))
            # print(t, x, normal_vector, surface_vector)
            # print(t, 'x- top')
            if self.check_if_ball_leaving(x_v, normal_vector):
                collided = True
        if x[2] < BALL_RADIUS:
            # floor
            normal_vector = np.array([0, 0, 1])
            if self.check_if_ball_leaving(x_v, normal_vector):
                collided = True
        elif x[2] > CEILING_DISTANCE - BALL_RADIUS:
            # ceiling
            normal_vector = np.array([0, 0, -1])
            if self.check_if_ball_leaving(x_v, normal_vector):
                collided = True
        # sides
        if x[0] < -SIDE_WALL_DISTANCE + BALL_RADIUS:
            normal_vector = np.array([1, 0, 0])
            if self.check_if_ball_leaving(x_v, normal_vector):
                collided = True
        elif x[0] > SIDE_WALL_DISTANCE - BALL_RADIUS:
            normal_vector = np.array([-1, 0, 0])
            if self.check_if_ball_leaving(x_v, normal_vector):
                collided = True
        # back
        if x[1] < -BACK_WALL_DISTANCE + BALL_RADIUS and \
                BALL_RADIUS < x[2] < CEILING_DISTANCE - BALL_RADIUS and \
                (abs(x[0]) > GOAL_X - BALL_RADIUS or abs(x[2]) > GOAL_Z - BALL_RADIUS):
            normal_vector = np.array([0, 1, 0])
            if self.check_if_ball_leaving(x_v, normal_vector):
                collided = True
        elif x[1] > BACK_WALL_DISTANCE - BALL_RADIUS and \
                BALL_RADIUS < x[2] < CEILING_DISTANCE - BALL_RADIUS and \
                (abs(x[0]) > GOAL_X - BALL_RADIUS or abs(x[2]) > GOAL_Z - BALL_RADIUS):
            normal_vector = np.array([0, -1, 0])
            if self.check_if_ball_leaving(x_v, normal_vector):
                collided = True

        # corner side
        if abs(x[0]) + abs(x[1]) + BALL_RADIUS > CORNER_WALL_DISTANCE:
            over_rt2 = 1 / np.sqrt(2)
            if x[0] < 0 and x[1] < 0:
                normal_vector = np.array([over_rt2, over_rt2, 0])
                if self.check_if_ball_leaving(x_v, normal_vector):
                    collided = True
            elif x[0] < 0 and x[1] > 0:
                normal_vector = np.array([over_rt2, -over_rt2, 0])
                if self.check_if_ball_leaving(x_v, normal_vector):
                    collided = True
            elif x[0] > 0 and x[1] < 0:
                normal_vector = np.array([-over_rt2, over_rt2, 0])
                if self.check_if_ball_leaving(x_v, normal_vector):
                    collided = True
            elif x[0] > 0 and x[1] > 0:
                normal_vector = np.array([-over_rt2, -over_rt2, 0])
                if self.check_if_ball_leaving(x_v, normal_vector):
                    collided = True

        if collided:
            state = (v, self.sim_vars['ang_vel'])
            new_state = bounce(state, normal_vector)
            v, self.sim_vars['ang_vel'] = new_state
            x_v[3:] = v

        # if v > max speed: v = v
        if v.dot(v) > self.ball_max_speed ** 2:
            v = v / np.sqrt(v.dot(v)) * self.ball_max_speed

        # calculate a
        a = np.array([0, 0, -self.gravity]) - self.air_resistance * v

        # if ang_vel > max rotation: normalise to 6
        ang_vel = self.sim_vars['ang_vel']
        if ang_vel.dot(ang_vel) > self.ball_max_rotation_speed ** 2:
            ang_vel = ang_vel / np.sqrt(ang_vel.dot(ang_vel)) * self.ball_max_rotation_speed
            self.sim_vars['ang_vel'] = ang_vel

        return np.concatenate((v, a)).flatten()

    def check_if_ball_leaving(self, x_v, normal_vector):
        if normal_vector.dot(x_v[3:6]) < 0:
            return True
        else:
            return False



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/metadata/player_loadout.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/metadata/player_loadout.proto',
  package='api.metadata',
  serialized_pb=_b('\n!api/metadata/player_loadout.proto\x12\x0c\x61pi.metadata\"\xdd\x01\n\rPlayerLoadout\x12\x13\n\x0bload_out_id\x18\x01 \x01(\x05\x12\x0e\n\x06\x62\x61nner\x18\x02 \x01(\x05\x12\r\n\x05\x62oost\x18\x03 \x01(\x05\x12\x0b\n\x03\x63\x61r\x18\x04 \x01(\x05\x12\x16\n\x0egoal_explosion\x18\x05 \x01(\x05\x12\x0c\n\x04skin\x18\x06 \x01(\x05\x12\r\n\x05trail\x18\x07 \x01(\x05\x12\x0e\n\x06wheels\x18\x08 \x01(\x05\x12\x0f\n\x07version\x18\t \x01(\x05\x12\x0e\n\x06topper\x18\n \x01(\x05\x12\x0f\n\x07\x61ntenna\x18\x0b \x01(\x05\x12\x14\n\x0c\x65ngine_audio\x18\x0c \x01(\x05')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_PLAYERLOADOUT = _descriptor.Descriptor(
  name='PlayerLoadout',
  full_name='api.metadata.PlayerLoadout',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='load_out_id', full_name='api.metadata.PlayerLoadout.load_out_id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='banner', full_name='api.metadata.PlayerLoadout.banner', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='boost', full_name='api.metadata.PlayerLoadout.boost', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='car', full_name='api.metadata.PlayerLoadout.car', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='goal_explosion', full_name='api.metadata.PlayerLoadout.goal_explosion', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='skin', full_name='api.metadata.PlayerLoadout.skin', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='trail', full_name='api.metadata.PlayerLoadout.trail', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='wheels', full_name='api.metadata.PlayerLoadout.wheels', index=7,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='version', full_name='api.metadata.PlayerLoadout.version', index=8,
      number=9, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='topper', full_name='api.metadata.PlayerLoadout.topper', index=9,
      number=10, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='antenna', full_name='api.metadata.PlayerLoadout.antenna', index=10,
      number=11, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='engine_audio', full_name='api.metadata.PlayerLoadout.engine_audio', index=11,
      number=12, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=52,
  serialized_end=273,
)

DESCRIPTOR.message_types_by_name['PlayerLoadout'] = _PLAYERLOADOUT

PlayerLoadout = _reflection.GeneratedProtocolMessageType('PlayerLoadout', (_message.Message,), dict(
  DESCRIPTOR = _PLAYERLOADOUT,
  __module__ = 'api.metadata.player_loadout_pb2'
  # @@protoc_insertion_point(class_scope:api.metadata.PlayerLoadout)
  ))
_sym_db.RegisterMessage(PlayerLoadout)


# @@protoc_insertion_point(module_scope)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/stats/events.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ...api import player_id_pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/stats/events.proto',
  package='api.stats',
  serialized_pb=_b('\n\x16\x61pi/stats/events.proto\x12\tapi.stats\x1a\x13\x61pi/player_id.proto\"7\n\x08\x42\x61llData\x12\r\n\x05pos_x\x18\x01 \x01(\x02\x12\r\n\x05pos_y\x18\x02 \x01(\x02\x12\r\n\x05pos_z\x18\x03 \x01(\x02\"\xfb\x03\n\x03Hit\x12\x14\n\x0c\x66rame_number\x18\x01 \x01(\x05\x12 \n\tplayer_id\x18\x02 \x01(\x0b\x32\r.api.PlayerId\x12\x1a\n\x12\x63ollision_distance\x18\x03 \x01(\x02\x12&\n\tball_data\x18\x04 \x01(\x0b\x32\x13.api.stats.BallData\x12\r\n\x05pass_\x18\x05 \x01(\x08\x12\x0e\n\x06passed\x18\x06 \x01(\x08\x12\x0f\n\x07\x64ribble\x18\x07 \x01(\x08\x12\x1c\n\x14\x64ribble_continuation\x18\x08 \x01(\x08\x12\x0c\n\x04shot\x18\t \x01(\x08\x12\x0c\n\x04goal\x18\n \x01(\x08\x12\x0e\n\x06\x61ssist\x18\x0c \x01(\x08\x12\x10\n\x08\x61ssisted\x18\r \x01(\x08\x12\x0c\n\x04save\x18\x0e \x01(\x08\x12\x0e\n\x06\x61\x65rial\x18\x0f \x01(\x08\x12\x11\n\ton_ground\x18\x10 \x01(\x08\x12\x10\n\x08\x64istance\x18\x11 \x01(\x02\x12\x18\n\x10\x64istance_to_goal\x18\x12 \x01(\x02\x12!\n\x19previous_hit_frame_number\x18\x13 \x01(\x05\x12\x1d\n\x15next_hit_frame_number\x18\x14 \x01(\x05\x12\x13\n\x0bgoal_number\x18\x15 \x01(\x05\x12\x12\n\nis_kickoff\x18\x16 \x01(\x08\x12$\n\x1c\x44\x45PRECATED_field_goal_number\x18\x0b \x01(\x08\"s\n\x04\x42ump\x12\x14\n\x0c\x66rame_number\x18\x01 \x01(\x05\x12\"\n\x0b\x61ttacker_id\x18\x02 \x01(\x0b\x32\r.api.PlayerId\x12 \n\tvictim_id\x18\x03 \x01(\x0b\x32\r.api.PlayerId\x12\x0f\n\x07is_demo\x18\x04 \x01(\x08\"\xb1\x01\n\x0c\x43\x61meraChange\x12\x14\n\x0c\x66rame_number\x18\x01 \x01(\x05\x12 \n\tplayer_id\x18\x02 \x01(\x0b\x32\r.api.PlayerId\x12\x1a\n\x12\x64istance_from_ball\x18\x03 \x01(\x02\x12\x18\n\x10is_in_possession\x18\x04 \x01(\x08\x12\x1b\n\x13length_in_this_mode\x18\x05 \x01(\x02\x12\x16\n\x0eis_on_ball_cam\x18\x06 \x01(\x08\"}\n\nFiftyFifty\x12\x14\n\x0c\x66rame_number\x18\x01 \x01(\x05\x12\x1e\n\x07players\x18\x02 \x03(\x0b\x32\r.api.PlayerId\x12\x12\n\nis_kickoff\x18\x03 \x01(\x08\x12%\n\x0ewinning_player\x18\x04 \x01(\x0b\x32\r.api.PlayerId\"?\n\x07Kickoff\x12\x1a\n\x12start_frame_number\x18\x01 \x01(\x05\x12\x18\n\x10\x65nd_frame_number\x18\x02 \x01(\x05')
  ,
  dependencies=[player_id_pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_BALLDATA = _descriptor.Descriptor(
  name='BallData',
  full_name='api.stats.BallData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pos_x', full_name='api.stats.BallData.pos_x', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pos_y', full_name='api.stats.BallData.pos_y', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pos_z', full_name='api.stats.BallData.pos_z', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=58,
  serialized_end=113,
)


_HIT = _descriptor.Descriptor(
  name='Hit',
  full_name='api.stats.Hit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='frame_number', full_name='api.stats.Hit.frame_number', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='player_id', full_name='api.stats.Hit.player_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='collision_distance', full_name='api.stats.Hit.collision_distance', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ball_data', full_name='api.stats.Hit.ball_data', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pass_', full_name='api.stats.Hit.pass_', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='passed', full_name='api.stats.Hit.passed', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dribble', full_name='api.stats.Hit.dribble', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dribble_continuation', full_name='api.stats.Hit.dribble_continuation', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='shot', full_name='api.stats.Hit.shot', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='goal', full_name='api.stats.Hit.goal', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='assist', full_name='api.stats.Hit.assist', index=10,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='assisted', full_name='api.stats.Hit.assisted', index=11,
      number=13, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='save', full_name='api.stats.Hit.save', index=12,
      number=14, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='aerial', full_name='api.stats.Hit.aerial', index=13,
      number=15, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='on_ground', full_name='api.stats.Hit.on_ground', index=14,
      number=16, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='distance', full_name='api.stats.Hit.distance', index=15,
      number=17, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='distance_to_goal', full_name='api.stats.Hit.distance_to_goal', index=16,
      number=18, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='previous_hit_frame_number', full_name='api.stats.Hit.previous_hit_frame_number', index=17,
      number=19, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='next_hit_frame_number', full_name='api.stats.Hit.next_hit_frame_number', index=18,
      number=20, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='goal_number', full_name='api.stats.Hit.goal_number', index=19,
      number=21, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='is_kickoff', full_name='api.stats.Hit.is_kickoff', index=20,
      number=22, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='DEPRECATED_field_goal_number', full_name='api.stats.Hit.DEPRECATED_field_goal_number', index=21,
      number=11, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=116,
  serialized_end=623,
)


_BUMP = _descriptor.Descriptor(
  name='Bump',
  full_name='api.stats.Bump',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='frame_number', full_name='api.stats.Bump.frame_number', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='attacker_id', full_name='api.stats.Bump.attacker_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='victim_id', full_name='api.stats.Bump.victim_id', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='is_demo', full_name='api.stats.Bump.is_demo', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=625,
  serialized_end=740,
)


_CAMERACHANGE = _descriptor.Descriptor(
  name='CameraChange',
  full_name='api.stats.CameraChange',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='frame_number', full_name='api.stats.CameraChange.frame_number', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='player_id', full_name='api.stats.CameraChange.player_id', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='distance_from_ball', full_name='api.stats.CameraChange.distance_from_ball', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='is_in_possession', full_name='api.stats.CameraChange.is_in_possession', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='length_in_this_mode', full_name='api.stats.CameraChange.length_in_this_mode', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='is_on_ball_cam', full_name='api.stats.CameraChange.is_on_ball_cam', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=743,
  serialized_end=920,
)


_FIFTYFIFTY = _descriptor.Descriptor(
  name='FiftyFifty',
  full_name='api.stats.FiftyFifty',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='frame_number', full_name='api.stats.FiftyFifty.frame_number', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='players', full_name='api.stats.FiftyFifty.players', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='is_kickoff', full_name='api.stats.FiftyFifty.is_kickoff', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='winning_player', full_name='api.stats.FiftyFifty.winning_player', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=922,
  serialized_end=1047,
)


_KICKOFF = _descriptor.Descriptor(
  name='Kickoff',
  full_name='api.stats.Kickoff',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='start_frame_number', full_name='api.stats.Kickoff.start_frame_number', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='end_frame_number', full_name='api.stats.Kickoff.end_frame_number', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1049,
  serialized_end=1112,
)

_HIT.fields_by_name['player_id'].message_type = player_id_pb2._PLAYERID
_HIT.fields_by_name['ball_data'].message_type = _BALLDATA
_BUMP.fields_by_name['attacker_id'].message_type = player_id_pb2._PLAYERID
_BUMP.fields_by_name['victim_id'].message_type = player_id_pb2._PLAYERID
_CAMERACHANGE.fields_by_name['player_id'].message_type = player_id_pb2._PLAYERID
_FIFTYFIFTY.fields_by_name['players'].message_type = player_id_pb2._PLAYERID
_FIFTYFIFTY.fields_by_name['winning_player'].message_type = player_id_pb2._PLAYERID
DESCRIPTOR.message_types_by_name['BallData'] = _BALLDATA
DESCRIPTOR.message_types_by_name['Hit'] = _HIT
DESCRIPTOR.message_types_by_name['Bump'] = _BUMP
DESCRIPTOR.message_types_by_name['CameraChange'] = _CAMERACHANGE
DESCRIPTOR.message_types_by_name['FiftyFifty'] = _FIFTYFIFTY
DESCRIPTOR.message_types_by_name['Kickoff'] = _KICKOFF

BallData = _reflection.GeneratedProtocolMessageType('BallData', (_message.Message,), dict(
  DESCRIPTOR = _BALLDATA,
  __module__ = 'api.stats.events_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.BallData)
  ))
_sym_db.RegisterMessage(BallData)

Hit = _reflection.GeneratedProtocolMessageType('Hit', (_message.Message,), dict(
  DESCRIPTOR = _HIT,
  __module__ = 'api.stats.events_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.Hit)
  ))
_sym_db.RegisterMessage(Hit)

Bump = _reflection.GeneratedProtocolMessageType('Bump', (_message.Message,), dict(
  DESCRIPTOR = _BUMP,
  __module__ = 'api.stats.events_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.Bump)
  ))
_sym_db.RegisterMessage(Bump)

CameraChange = _reflection.GeneratedProtocolMessageType('CameraChange', (_message.Message,), dict(
  DESCRIPTOR = _CAMERACHANGE,
  __module__ = 'api.stats.events_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.CameraChange)
  ))
_sym_db.RegisterMessage(CameraChange)

FiftyFifty = _reflection.GeneratedProtocolMessageType('FiftyFifty', (_message.Message,), dict(
  DESCRIPTOR = _FIFTYFIFTY,
  __module__ = 'api.stats.events_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.FiftyFifty)
  ))
_sym_db.RegisterMessage(FiftyFifty)

Kickoff = _reflection.GeneratedProtocolMessageType('Kickoff', (_message.Message,), dict(
  DESCRIPTOR = _KICKOFF,
  __module__ = 'api.stats.events_pb2'
  # @@protoc_insertion_point(class_scope:api.stats.Kickoff)
  ))
_sym_db.RegisterMessage(Kickoff)


# @@protoc_insertion_point(module_scope)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/metadata/camera_settings.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/metadata/camera_settings.proto',
  package='api.metadata',
  serialized_pb=_b('\n\"api/metadata/camera_settings.proto\x12\x0c\x61pi.metadata\"\xa7\x01\n\x0e\x43\x61meraSettings\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x11\n\tstiffness\x18\x02 \x01(\x02\x12\x0e\n\x06height\x18\x03 \x01(\x02\x12\x18\n\x10transition_speed\x18\x04 \x01(\x02\x12\r\n\x05pitch\x18\x05 \x01(\x02\x12\x14\n\x0cswivel_speed\x18\x06 \x01(\x02\x12\x15\n\rfield_of_view\x18\x07 \x01(\x02\x12\x10\n\x08\x64istance\x18\x08 \x01(\x02')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_CAMERASETTINGS = _descriptor.Descriptor(
  name='CameraSettings',
  full_name='api.metadata.CameraSettings',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='api.metadata.CameraSettings.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stiffness', full_name='api.metadata.CameraSettings.stiffness', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='height', full_name='api.metadata.CameraSettings.height', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='transition_speed', full_name='api.metadata.CameraSettings.transition_speed', index=3,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pitch', full_name='api.metadata.CameraSettings.pitch', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='swivel_speed', full_name='api.metadata.CameraSettings.swivel_speed', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='field_of_view', full_name='api.metadata.CameraSettings.field_of_view', index=6,
      number=7, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='distance', full_name='api.metadata.CameraSettings.distance', index=7,
      number=8, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=53,
  serialized_end=220,
)

DESCRIPTOR.message_types_by_name['CameraSettings'] = _CAMERASETTINGS

CameraSettings = _reflection.GeneratedProtocolMessageType('CameraSettings', (_message.Message,), dict(
  DESCRIPTOR = _CAMERASETTINGS,
  __module__ = 'api.metadata.camera_settings_pb2'
  # @@protoc_insertion_point(class_scope:api.metadata.CameraSettings)
  ))
_sym_db.RegisterMessage(CameraSettings)


# @@protoc_insertion_point(module_scope)



import os
import unittest

from carball.rattletrap.rattletrap_utils import get_rattletrap_binaries, get_rattletrap_path, download_rattletrap
from carball.rattletrap.run_rattletrap import create_rattletrap_command
from carball.rattletrap.check_rattletrap_version import update_rattletrap


class setup_tests(unittest.TestCase):

    def cleanup(self):
        path = get_rattletrap_path()
        binaries = get_rattletrap_binaries(get_rattletrap_path())

        # clean out existing
        for binary in binaries:
            os.remove(os.path.join(path, binary))

    def test_rattle(self):
        self.cleanup()

        download_rattletrap()

    def test_create_rattletrap_command(self):
        self.cleanup()

        create_rattletrap_command('file.replay', 'outputdir')

        # check post download
        create_rattletrap_command('file.replay', 'outputdir2')

    def test_direct_check_version(self):
        self.cleanup()
        update_rattletrap()

        #skip update
        update_rattletrap()

    def test_get_correct_version_from_platform(self):
        path = get_rattletrap_path()
        binaries = get_rattletrap_binaries(get_rattletrap_path())



class ApiPlayerLoadout:

    @staticmethod
    def create_from_player(proto_loadout, player):
        try:
            loadout = player.loadout[player.is_orange]
        except IndexError:
            return

        if loadout['banner'] is not None:
            proto_loadout.banner = loadout['banner']
        if loadout['boost'] is not None:
            proto_loadout.boost = loadout['boost']
        if loadout['car'] is not None:
            proto_loadout.car = loadout['car']
        if loadout['goal_explosion'] is not None:
            proto_loadout.goal_explosion = loadout['goal_explosion']
        if loadout['skin'] is not None:
            proto_loadout.skin = loadout['skin']
        if loadout['trail'] is not None:
            proto_loadout.trail = loadout['trail']
        if loadout['version'] is not None:
            proto_loadout.version = loadout['version']
        if loadout['wheels'] is not None:
            proto_loadout.wheels = loadout['wheels']
        if loadout['topper'] is not None:
            proto_loadout.topper = loadout['topper']
        if loadout['antenna'] is not None:
            proto_loadout.antenna = loadout['antenna']
        if loadout['engine_audio'] is not None:
            proto_loadout.engine_audio = loadout['engine_audio']



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/player.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..api.metadata import player_loadout_pb2
from ..api.metadata import camera_settings_pb2
from ..api.stats import player_stats_pb2
from ..api import player_id_pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/player.proto',
  package='api',
  serialized_pb=_b('\n\x10\x61pi/player.proto\x12\x03\x61pi\x1a!api/metadata/player_loadout.proto\x1a\"api/metadata/camera_settings.proto\x1a\x1c\x61pi/stats/player_stats.proto\x1a\x13\x61pi/player_id.proto\"\x97\x03\n\x06Player\x12\x19\n\x02id\x18\x01 \x01(\x0b\x32\r.api.PlayerId\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x10\n\x08title_id\x18\x03 \x01(\x05\x12\r\n\x05score\x18\x04 \x01(\x05\x12\r\n\x05goals\x18\x05 \x01(\x05\x12\x0f\n\x07\x61ssists\x18\x06 \x01(\x05\x12\r\n\x05saves\x18\x07 \x01(\x05\x12\r\n\x05shots\x18\x08 \x01(\x05\x12\x35\n\x0f\x63\x61mera_settings\x18\t \x01(\x0b\x32\x1c.api.metadata.CameraSettings\x12,\n\x07loadout\x18\n \x01(\x0b\x32\x1b.api.metadata.PlayerLoadout\x12\x11\n\tis_orange\x18\x0b \x01(\x05\x12%\n\x05stats\x18\x0c \x01(\x0b\x32\x16.api.stats.PlayerStats\x12#\n\x0cparty_leader\x18\r \x01(\x0b\x32\r.api.PlayerId\x12\x0e\n\x06is_bot\x18\x0e \x01(\x08\x12\x14\n\x0ctime_in_game\x18\x0f \x01(\x02\x12\x1b\n\x13\x66irst_frame_in_game\x18\x10 \x01(\x05')
  ,
  dependencies=[player_loadout_pb2.DESCRIPTOR,camera_settings_pb2.DESCRIPTOR,player_stats_pb2.DESCRIPTOR,player_id_pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_PLAYER = _descriptor.Descriptor(
  name='Player',
  full_name='api.Player',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='api.Player.id', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='api.Player.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='title_id', full_name='api.Player.title_id', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='score', full_name='api.Player.score', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='goals', full_name='api.Player.goals', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='assists', full_name='api.Player.assists', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='saves', full_name='api.Player.saves', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='shots', full_name='api.Player.shots', index=7,
      number=8, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='camera_settings', full_name='api.Player.camera_settings', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='loadout', full_name='api.Player.loadout', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='is_orange', full_name='api.Player.is_orange', index=10,
      number=11, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='stats', full_name='api.Player.stats', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='party_leader', full_name='api.Player.party_leader', index=12,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='is_bot', full_name='api.Player.is_bot', index=13,
      number=14, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time_in_game', full_name='api.Player.time_in_game', index=14,
      number=15, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='first_frame_in_game', full_name='api.Player.first_frame_in_game', index=15,
      number=16, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=148,
  serialized_end=555,
)

_PLAYER.fields_by_name['id'].message_type = player_id_pb2._PLAYERID
_PLAYER.fields_by_name['camera_settings'].message_type = camera_settings_pb2._CAMERASETTINGS
_PLAYER.fields_by_name['loadout'].message_type = player_loadout_pb2._PLAYERLOADOUT
_PLAYER.fields_by_name['stats'].message_type = player_stats_pb2._PLAYERSTATS
_PLAYER.fields_by_name['party_leader'].message_type = player_id_pb2._PLAYERID
DESCRIPTOR.message_types_by_name['Player'] = _PLAYER

Player = _reflection.GeneratedProtocolMessageType('Player', (_message.Message,), dict(
  DESCRIPTOR = _PLAYER,
  __module__ = 'api.player_pb2'
  # @@protoc_insertion_point(class_scope:api.Player)
  ))
_sym_db.RegisterMessage(Player)


# @@protoc_insertion_point(module_scope)



import logging

import pandas as pd

from ...json_parser.game import Game
logger = logging.getLogger(__name__)


class SaltieGame:

    @staticmethod
    def get_first_touch_frames(game):
        if game.frames.loc[:, 'ball_has_been_hit'].any():
            ball_has_been_hit = game.frames.loc[:, 'ball_has_been_hit']
            last_frame_ball_has_been_hit = ball_has_been_hit.shift(1).rename('last_frame_ball_has_been_hit')
            ball_hit_dataframe = pd.concat([ball_has_been_hit, last_frame_ball_has_been_hit], axis=1)
            ball_hit_dataframe.fillna(False, inplace=True)

            kickoff_frames = ball_hit_dataframe[(ball_hit_dataframe['ball_has_been_hit']) &
                                                ~(ball_hit_dataframe['last_frame_ball_has_been_hit'])]
        else:
            logger.debug("No ball_has_been_hit?! Is this really old or what.")
            hit_team_no = game.ball.loc[:, 'hit_team_no']
            last_hit_team_no = hit_team_no.shift(1).rename('last_hit_team_no')
            hit_team_no_dataframe = pd.concat([hit_team_no, last_hit_team_no], axis=1)
            kickoff_frames = hit_team_no_dataframe[~(hit_team_no_dataframe['hit_team_no'].isnull()) &
                                                   (hit_team_no_dataframe['last_hit_team_no'].isnull())]

        return kickoff_frames.index.values

    @staticmethod
    def get_kickoff_frames(game):
        if game.frames.loc[:, 'replicated_seconds_remaining'].any():
            ball_has_been_hit = game.frames.loc[:, 'replicated_seconds_remaining']
            last_frame_ball_has_been_hit = ball_has_been_hit.shift(1).rename('last_replicated_seconds_remaining')
            ball_hit_dataframe = pd.concat([ball_has_been_hit, last_frame_ball_has_been_hit], axis=1)
            ball_hit_dataframe.fillna(0, inplace=True)

            kickoff_frames = ball_hit_dataframe[(ball_hit_dataframe['replicated_seconds_remaining'] == 0) &
                                                (ball_hit_dataframe['last_replicated_seconds_remaining'] > 0)]
        else:
            logger.debug("No ball_has_been_hit?! Is this really old or what.")
            hit_team_no = game.ball.loc[:, 'hit_team_no']
            last_hit_team_no = hit_team_no.shift(1).rename('last_hit_team_no')
            hit_team_no_dataframe = pd.concat([hit_team_no, last_hit_team_no], axis=1)
            kickoff_frames = hit_team_no_dataframe[~(hit_team_no_dataframe['hit_team_no'].isnull()) &
                                                   (hit_team_no_dataframe['last_hit_team_no'].isnull())]

        return kickoff_frames.index.values

    @staticmethod
    def create_data_df(game: Game) -> pd.DataFrame:
        data_dict = {player.name: player.data for player in game.players}
        data_dict['ball'] = game.ball
        initial_df = pd.concat(data_dict, axis=1)

        data_frame = pd.concat([initial_df, game.frames], axis=1)
        cols = []
        for c in data_frame.columns.values:
            if isinstance(c, str):
                cols.append(('game', c))
            else:
                cols.append(c)
        data_frame.columns = pd.MultiIndex.from_tuples(cols)
        return data_frame



from typing import List

from carball.generated.api import game_pb2
from carball.json_parser.game import Game

from carball.analysis.analysis_manager import AnalysisManager


class PerGoalAnalysis(AnalysisManager):

    def __init__(self, game: Game):
        super().__init__(game)
        self.protobuf_games = []

    def perform_full_analysis(self, game: Game, proto_game: game_pb2.Game, player_map, data_frame, kickoff_frames):
        self.protobuf_games = []
        # split up frames
        total_score = proto_game.game_metadata.score.team_0_score + proto_game.game_metadata.score.team_1_score
        print(total_score)
        for i in range(0, total_score):
            split_pandas = data_frame[data_frame.game.goal_number == i]
            new_proto = game_pb2.Game()
            for team in proto_game.teams:
                proto_team = new_proto.teams.add()
                proto_team.CopyFrom(team)

            for player in proto_game.players:
                proto_player = new_proto.players.add()
                proto_player.CopyFrom(player)
            new_proto.game_metadata.CopyFrom(proto_game.game_metadata)
            new_game = Game()
            new_game.players = game.players
            new_game.teams = game.teams
            new_game.frames = split_pandas
            super().perform_full_analysis(new_game, new_proto, player_map, split_pandas, kickoff_frames)
            self.protobuf_games.append(new_proto)

    def get_protobuf_data(self) -> List[game_pb2.Game]:
        """
        :return: The protobuf data created by the analysis
        """
        return self.protobuf_games



def initalize_rattletrap():
    from carball.rattletrap.check_rattletrap_version import update_rattletrap
    try:
        update_rattletrap()
    except e:
        print('Issue adding rattletrap')

def initialize_project():
    from utils.create_proto import create_proto_files
    from utils.import_fixer import convert_to_relative_imports
    
    create_proto_files()
    convert_to_relative_imports()
    initalize_rattletrap()

if __name__ == "__main__":
    initialize_project()



from typing import Dict

import pandas as pd

from carball.analysis.constants.basic_math import get_player_ball_displacements, get_distance_from_displacements, \
    get_position_displacements
from ....analysis.stats.utils.pandas_utils import sum_deltas_by_player_name
from ....analysis.stats.stats import BaseStat
from ....generated.api import game_pb2
from ....generated.api.player_pb2 import Player
from ....generated.api.stats.player_stats_pb2 import PlayerStats
from ....json_parser.game import Game


class BallDistanceStat(BaseStat):

    def calculate_player_stat(self, player_stat_map: Dict[str, PlayerStats], game: Game, proto_game: game_pb2.Game,
                              player_map: Dict[str, Player], data_frame: pd.DataFrame):

        player_distances_data_frame, player_distance_times, player_distance_with_delta, \
             = self.calculate_player_distance_to_location(player_map, data_frame,
                                                          data_frame['ball'])

        for player_id in player_stat_map.keys():
            close_frames = player_distances_data_frame[player_id] < 500
            time_close_to_ball = player_distance_with_delta[close_frames]['delta'].sum()
            distance_stats = player_stat_map[player_id].distance

            distance_stats.time_close_to_ball = time_close_to_ball

            try:
                distance_stats.time_closest_to_ball = player_distance_times['closest_player'][player_id]
                distance_stats.time_furthest_from_ball = player_distance_times['furthest_player'][player_id]
            except (AttributeError, KeyError):
                distance_stats.time_closest_to_ball = 0
                distance_stats.time_furthest_from_ball = 0

    @staticmethod
    def calculate_player_distance_to_location(player_map: Dict[str, Player], data_frame: pd.DataFrame,
                                              location: pd.DataFrame) -> (pd.DataFrame, pd.DataFrame, pd.DataFrame):
        """
        Calculates the player distance to a particular location
        also grabbing the closest and furthest player to those locations.
        :param player_map:
        :param data_frame:
        :param location:
        :return:
        """
        player_displacements = {player_id: get_position_displacements(data_frame[player_map[player_id].name],
                                                                      location)
                                for player_id in player_map.keys()}

        player_distances = {player_id: get_distance_from_displacements(player_data_frame).rename(player_id)
                            for player_id, player_data_frame in player_displacements.items()}

        player_distances_data_frame = pd.concat(player_distances, axis=1)
        closest_players = player_distances_data_frame.idxmin(axis=1).rename('closest_player')
        furthest_players = player_distances_data_frame.idxmax(axis=1).rename('furthest_player')
        player_distance_with_delta = pd.concat([player_distances_data_frame, data_frame['game', 'delta'].rename('delta')], axis=1)

        player_ball_distance_times = pd.concat([
            sum_deltas_by_player_name(data_frame, players_data_frame)
            for players_data_frame in [closest_players, furthest_players]
        ], axis=1)

        player_ball_distance_times.fillna(value=0, inplace=True)
        return player_distances_data_frame, player_ball_distance_times, player_distance_with_delta



import logging
from typing import Dict

import pandas as pd

from ...generated.api import game_pb2
from ...generated.api.player_pb2 import Player
from ...generated.api.stats.events_pb2 import Hit
from ...generated.api.stats.player_stats_pb2 import PlayerStats
from ...generated.api.stats.team_stats_pb2 import TeamStats
from ...json_parser.game import Game


class BaseStat:

    def __init__(self):
        self.logger = logging.getLogger(type(self).__name__)

    def calculate_stat(self, proto_stat, game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                       data_frame: pd.DataFrame):
        """
        Calculates stats that applies to the general game or applies to players + teams at the same time.
        :param proto_stat: This is protobuf object for general game stats
        :param game: The raw data that has been created from python.
        :param proto_game: A protobuf that contains some parsed stats + all metadata for the game.
        :param player_map: A map of playerId to the protobuf Player object
        :param data_frame: The raw frames of the replay this is the same object as `game.frames`
        """
        raise NotImplementedError()

    def calculate_player_stat(self, player_stat_map: Dict[str, PlayerStats], game: Game, proto_game: game_pb2.Game,
                              player_map: Dict[str, Player], data_frame: pd.DataFrame):
        """
        Calculates stats that only apply to players.
        :param player_stat_map: A map of playerId to the specific proto object for stats for that player.
        :param game: The raw data that has been created from python.
        :param proto_game: A protobuf that contains some parsed stats + all metadata for the game.
        :param player_map: A map of playerId to the protobuf Player object
        :param data_frame: The raw frames of the replay this is the same object as `game.frames`
        """
        raise NotImplementedError()

    def calculate_team_stat(self, team_stat_list: Dict[int, TeamStats], game: Game, proto_game: game_pb2.Game,
                            player_map: Dict[str, Player], data_frame: pd.DataFrame):
        """
        Calculate stats that only applies to teams
        :param team_stat_list: I map of team id to the specific proto object for stats for that team
        :param game: The raw data that has been created from python.
        :param proto_game: A protobuf that contains some parsed stats + all metadata for the game.
        :param player_map: A map of playerId to the protobuf Player object
        :param data_frame: The raw frames of the replay this is the same object as `game.frames`
        """
        raise NotImplementedError()


class HitStat:
    def initialize_hit_stat(self, game: Game, player_map: Dict[str, Player], data_frame: pd.DataFrame):
        """
        Called only once at the beginning of stat creation.
        :param game: The raw data that has been created from python.
        :param player_map: A map of playerId to the protobuf Player object
        :param data_frame: The raw frames of the replay this is the same object as `game.frames`
        """
        raise NotImplementedError()

    def calculate_next_hit_stat(self, game: Game, proto_game: game_pb2.Game, saltie_hit: Hit, next_saltie_hit: Hit,
                                player_map: Dict[str, Player], hit_index: int):
        """
        Calculate stats that use only the current hit + the next hit.
        :param game: The raw data that has been created from python.
        :param proto_game: A protobuf that contains some parsed stats + all metadata for the game.
        :param saltie_hit: The current hit we are looking at, this is a protobuf object.
        :param next_saltie_hit:  The hit that occured after the current one.
        :param player_map: A map of playerId to the protobuf Player object
        :param hit_index: The index in the list of protobuf hits where the current hit is listed.
        """
        raise NotImplementedError()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: api/game.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..api.metadata import game_metadata_pb2
from ..api.metadata import mutators_pb2
from ..api.stats import game_stats_pb2
from ..api import player_pb2
from ..api import team_pb2
from ..api import party_pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='api/game.proto',
  package='api',
  serialized_pb=_b('\n\x0e\x61pi/game.proto\x12\x03\x61pi\x1a api/metadata/game_metadata.proto\x1a\x1b\x61pi/metadata/mutators.proto\x1a\x1a\x61pi/stats/game_stats.proto\x1a\x10\x61pi/player.proto\x1a\x0e\x61pi/team.proto\x1a\x0f\x61pi/party.proto\"\xfa\x01\n\x04Game\x12\x31\n\rgame_metadata\x18\x01 \x01(\x0b\x32\x1a.api.metadata.GameMetadata\x12\x1c\n\x07players\x18\x02 \x03(\x0b\x32\x0b.api.Player\x12\x18\n\x05teams\x18\x03 \x03(\x0b\x32\t.api.Team\x12(\n\ngame_stats\x18\x04 \x01(\x0b\x32\x14.api.stats.GameStats\x12\x1b\n\x07parties\x18\x05 \x03(\x0b\x32\n.api.Party\x12\x0f\n\x07version\x18\x06 \x01(\x05\x12(\n\x08mutators\x18\x07 \x01(\x0b\x32\x16.api.metadata.Mutators*\x05\x08\x64\x10\xc8\x01')
  ,
  dependencies=[game_metadata_pb2.DESCRIPTOR,mutators_pb2.DESCRIPTOR,game_stats_pb2.DESCRIPTOR,player_pb2.DESCRIPTOR,team_pb2.DESCRIPTOR,party_pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_GAME = _descriptor.Descriptor(
  name='Game',
  full_name='api.Game',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='game_metadata', full_name='api.Game.game_metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='players', full_name='api.Game.players', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='teams', full_name='api.Game.teams', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='game_stats', full_name='api.Game.game_stats', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='parties', full_name='api.Game.parties', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='version', full_name='api.Game.version', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='mutators', full_name='api.Game.mutators', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=True,
  extension_ranges=[(100, 200), ],
  oneofs=[
  ],
  serialized_start=166,
  serialized_end=416,
)

_GAME.fields_by_name['game_metadata'].message_type = game_metadata_pb2._GAMEMETADATA
_GAME.fields_by_name['players'].message_type = player_pb2._PLAYER
_GAME.fields_by_name['teams'].message_type = team_pb2._TEAM
_GAME.fields_by_name['game_stats'].message_type = game_stats_pb2._GAMESTATS
_GAME.fields_by_name['parties'].message_type = party_pb2._PARTY
_GAME.fields_by_name['mutators'].message_type = mutators_pb2._MUTATORS
DESCRIPTOR.message_types_by_name['Game'] = _GAME

Game = _reflection.GeneratedProtocolMessageType('Game', (_message.Message,), dict(
  DESCRIPTOR = _GAME,
  __module__ = 'api.game_pb2'
  # @@protoc_insertion_point(class_scope:api.Game)
  ))
_sym_db.RegisterMessage(Game)


# @@protoc_insertion_point(module_scope)



import hashlib
import logging

bots = [
    "Armstrong",
    "Bandit",
    "Beast",
    "Boomer",
    "Buzz",
    "Casper",
    "Caveman",
    "C-Block",
    "Centice",
    "Chipper",
    "Cougar",
    "Dude",
    "Foamer",
    "Fury",
    "Gerwin",
    "Goose",
    "Heater",
    "Hollywood",
    "Hound",
    "Iceman",
    "Imp",
    "Jester",
    "JM",
    "Junker",
    "Khan",
    "Maverick",
    "Middy",
    "Merlin",
    "Mountain",
    "Myrtle",
    "Outlaw",
    "Poncho",
    "Rainmaker",
    "Raja",
    "Rex",
    "Roundhouse",
    "Sabretooth",
    "Saltie",
    "Samara",
    "Scout",
    "Shepard",
    "Slider",
    "Squall",
    "Sticks",
    "Stinger",
    "Storm",
    "Sundown",
    "Sultan",
    "Swabbie",
    "Tusk",
    "Tex",
    "Viper",
    "Wolfman",
    "Yuri",
]

logger = logging.getLogger(__name__)


def h11(w):
    return hashlib.md5(w).hexdigest()[:9]


def get_bot_map():
    result = dict()
    for i in range(len(bots)):
        result[bots[i]] = i + 1
    return result


def get_online_id_for_bot(bot_map, player):
    try:
        return 'b' + str(bot_map[player.name]) + 'b'
    except:
        logger.warning('Found bot not in bot list')
        try:
            return 'b' + h11(str(player.name).encode('utf-8').lower()) + 'b'
        except:
            logger.warning('bot has invalid name')
            return 'invalid_name'




from typing import Callable, Dict

import pandas as pd

from ....analysis.constants.field_constants import FieldConstants
from ....analysis.stats.stats import BaseStat
from ....generated.api import game_pb2
from ....generated.api.player_pb2 import Player
from ....generated.api.stats import stats_pb2
from ....generated.api.stats.player_stats_pb2 import PlayerStats
from ....json_parser.game import Game


class PositionalTendencies(BaseStat):
    field_constants = FieldConstants()

    def __init__(self):
        super().__init__()
        self.field_constants = FieldConstants()
        self.map_player_attributes_to_predicates = {
            "height_0": self.field_constants.get_height_0,
            "height_1": self.field_constants.get_height_1,
            "height_2": self.field_constants.get_height_2,
            "half_0": self.field_constants.get_half_0,
            "half_1": self.field_constants.get_half_1,
            "third_0": self.field_constants.get_third_0,
            "third_1": self.field_constants.get_third_1,
            "third_2": self.field_constants.get_third_2,
            "ball_0": self.field_constants.get_ball_0,
            "ball_1": self.field_constants.get_ball_1,
            "wall": self.field_constants.get_wall_time,
            "corner": self.field_constants.get_corner_time
        }

        self.map_ball_attributes_to_predicates = {
            "height_0": self.field_constants.get_height_0_ball,
            "height_1": self.field_constants.get_height_1,
            "height_2": self.field_constants.get_height_2,
            "half_0": self.field_constants.get_half_0,
            "half_1": self.field_constants.get_half_1,
            "third_0": self.field_constants.get_third_0,
            "third_1": self.field_constants.get_third_1,
            "third_2": self.field_constants.get_third_2,
            "ball_0": self.field_constants.get_ball_0,
            "ball_1": self.field_constants.get_ball_1,
            "wall": self.field_constants.get_wall_time,
            "corner": self.field_constants.get_corner_time
        }

    def calculate_player_stat(self, player_stat_map: Dict[str, PlayerStats], game: Game, proto_game: game_pb2.Game,
                              player_map: Dict[str, Player], data_frame: pd.DataFrame):
        for id, player in player_map.items():
            self.get_player_tendencies(player, data_frame)

    def calculate_stat(self, proto_stat, game: Game, proto_game: game_pb2.Game, player_map: Dict[str, Player],
                       data_frame: pd.DataFrame):
        ball_data_frame = data_frame['ball']

        self.get_tendencies(data_frame, ball_data_frame, ball_data_frame, False,
                            proto_stat.ball_stats.positional_tendencies, self.map_ball_attributes_to_predicates)

    def get_player_tendencies(self, player: Player, data_frame: pd.DataFrame):
        self.get_tendencies(data_frame, data_frame[player.name], data_frame['ball'],
                            player.is_orange, player.stats.positional_tendencies,
                            self.map_player_attributes_to_predicates)

    def get_tendencies(self, data_frame: pd.DataFrame, player_data_frame: pd.DataFrame,
                       ball_data_frame: pd.DataFrame, is_orange: bool,
                       positional_tendencies: stats_pb2.PositionalTendencies,
                       predicate_map: Dict[str, Callable]):
        """
        Calculates the tendencies for all of the stats
        :param data_frame: The normal data frame
        :param player_data_frame: The frame for the positions of the player
        :param ball_data_frame: The frame for the positions of the ball
        :param is_orange: If we need to flip the field for defense/offense
        :param positional_tendencies: The proto field where everything is set
        :param predicate_map: The map of checks to call
        """
        player_ball_dataframes: Dict[str, pd.DataFrame] = {
            "player_data_frame": player_data_frame,
            "ball_data_frame": ball_data_frame,
        }
        if is_orange:
            player_ball_dataframes = self.get_flipped_dataframes(player_ball_dataframes)

        init_params = {
            attr: self.get_duration_from_predicate(predicate, player_ball_dataframes, data_frame)
            for attr, predicate in predicate_map.items()
        }
        self.set_tendency_proto(positional_tendencies, **init_params)

    @staticmethod
    def get_duration_from_predicate(predicate: Callable,
                                    player_ball_dataframes: Dict[str, pd.DataFrame], data_frame: pd.DataFrame):
        boolean_index = predicate(**player_ball_dataframes)
        deltas = data_frame.game.delta
        return deltas[boolean_index].sum()

    def get_flipped_dataframes(self, data_frames: Dict[str, pd.DataFrame]) -> Dict[str, pd.DataFrame]:
        new_dataframes = {
            _k: _v.copy() for _k, _v in data_frames.items()
        }

        for data_frame in new_dataframes.values():
            data_frame.pos_y *= -1
            data_frame.rot_y += 65535 / 2
            data_frame.rot_y %= 65536

        return new_dataframes

    @staticmethod
    def set_tendency_proto(proto: stats_pb2.PositionalTendencies, height_0: float, height_1: float, height_2: float,
                           half_0: float, half_1: float,
                           third_0: float, third_1: float, third_2: float,
                           ball_0: float, ball_1: float,
                           wall: float, corner: float
                           ):
        """
        :param proto: What object everything is getting set on
        :param height_0: Time spent on ground
        :param height_1: Time spent low in air
        :param height_2: Time spent high in air
        :param half_0: Time spent in defending half
        :param half_1: Time spent in attacking half
        :param third_0: Time spent in defending third
        :param third_1: Time spent in middle third
        :param third_2: Time spent in attacking third
        :param ball_0: Time spent behind ball
        :param ball_1: Time spent ahead of ball
        :param wall: Time spent near the wall
        :param corner: Time spent near the corner
        """
        proto.time_on_ground = height_0
        proto.time_low_in_air = height_1
        proto.time_high_in_air = height_2
        proto.time_in_defending_half = half_0
        proto.time_in_attacking_half = half_1
        proto.time_in_defending_third = third_0
        proto.time_in_neutral_third = third_1
        proto.time_in_attacking_third = third_2
        proto.time_behind_ball = ball_0
        proto.time_in_front_ball = ball_1
        proto.time_near_wall = wall
        proto.time_in_corner = corner



import io
import logging

import pandas as pd
import numpy as np

from carball.analysis.constants.basic_math import positional_columns
from carball.analysis.utils.numpy_manager import write_array_to_file, read_array_from_file


logger = logging.getLogger(__name__)


class PandasManager:

    @staticmethod
    def safe_write_pandas_to_memory(df, field_name=""):
        try:
            # return PandasManager.write_hdf_to_buffer(df)
            return PandasManager.write_numpy_to_memory(df)
        except BaseException as e:
            logger.exception("Failure to write pandas [%s] to memory: %s", field_name, e)

    @staticmethod
    def safe_read_pandas_to_memory(buffer, field_name=""):
        try:
            return PandasManager.read_numpy_from_memory(buffer)
        except BaseException as e:
            logger.exception("Failure to read pandas [%s] from memory: %s", field_name, e)

    @staticmethod
    def write_numpy_to_memory(df):
        numpy_array = df.to_records(index=True)
        compressed_array = io.BytesIO()
        write_array_to_file(compressed_array, numpy_array)
        return compressed_array.getvalue()

    @staticmethod
    def read_numpy_from_memory(buffer):
        array = read_array_from_file(buffer)
        dataframe = pd.DataFrame.from_records(array)
        dataframe.set_index('index', drop=True, inplace=True)
        columns = []
        for tuple_str in dataframe.columns.values:
            columns.append(eval(tuple_str))
        dataframe.columns = pd.MultiIndex.from_tuples(columns)
        return dataframe

    @staticmethod
    def write_pandas_to_buffer_for_tooling(df, players):
        columns = []
        columns.append(df['ball'][positional_columns].values)
        for player in players:
            columns.append(df[player.name][positional_columns].values)
        numpy_array = np.concatenate(columns, axis=1)
        compressed_array = io.BytesIO()
        np.save(compressed_array, numpy_array, allow_pickle=True, fix_imports=False)
        return compressed_array.getvalue()
